@@ -11,7 +11,7 @@ class TAFError(Exception):
 class CloneRepoException(TAFError):
     def __init__(self, repo):
         self.message = (
-            f"Cannot clone {repo.name} from any of the following URLs: {repo.repo_urls}"
+            f"Cannot clone {repo.name} from any of the following URLs: {repo.urls}"
         )
 
 
@@ -52,6 +52,14 @@ class InvalidKeyError(TAFError):
         )
 
 
+class MissingHostsError(TAFError):
+    pass
+
+
+class InvalidHostsError(TAFError):
+    pass
+
+
 class InvalidOrMissingMetadataError(TAFError):
     pass
 
@@ -71,6 +79,16 @@ class RepositoryInstantiationError(TAFError):
         self.message = message
 
 
+class ScriptExecutionError(TAFError):
+    def __init__(self, script, error_msg):
+        message = (
+            f"An error happened during execution of script {script}:\n\n: {error_msg}"
+        )
+        super().__init__(message)
+        self.message = message
+        self.script = script
+
+
 class MetadataUpdateError(TAFError):
     def __init__(self, metadata_role, message):
         super().__init__(
@@ -129,5 +147,20 @@ class UpdateFailedError(TAFError):
     pass
 
 
+class UpdaterAdditionalCommitsError(TAFError):
+    def __init__(self, additional_commits_per_repo, message=None):
+        self.additional_commits_per_repo = additional_commits_per_repo
+        if message is None:
+            message = ""
+            for repo, branch_commits in additional_commits_per_repo.items():
+                for branch, commits in branch_commits.items():
+                    message += f"Repository {repo}: branch {branch} contains {len(commits)} additional"
+                    if len(commits) > 1:
+                        message += "commits\n"
+                    else:
+                        message += "commit\n"
+        self.message = message
+
+
 class YubikeyError(Exception):
     pass
