@@ -24,39 +24,82 @@ EMPTY_TREE = "4b825dc642cb6eb9a060e54bf8d69288fbee4904"
 class GitRepository:
     def __init__(
         self,
-        path,
-        repo_urls=None,
-        additional_info=None,
-        default_branch="master",
-        repo_name=None,
+        library_dir=None,
+        name=None,
+        urls=None,
+        custom=None,
+        default_branch="main",
+        path=None,
         *args,
         **kwargs,
     ):
         """
-    Args:
-      path: repository's path
-      repo_urls: repository's urls (optional)
-      additional_info: a dictionary containing other data (optional)
-      default_branch: repository's default branch
-    """
-        self._path = Path(path).resolve()
-        if repo_name is None:
-            repo_name = self._path.name
-        self.name = repo_name
-        self.default_branch = default_branch
-        if repo_urls is not None:
-            if settings.update_from_filesystem is False:
-                for url in repo_urls:
-                    self._validate_url(url)
+        Args:
+          library_dir (Path or str): path to the library root. This is a directory which contains other repositories,
+          whose full path is determined by appending their names to the library dir. A repository can be
+          instantiated by specifying library dir and name, or by the full path.
+          name (str): repository's name, which is appended to the library dir to form the full path.
+          Must be in the namespace/name format. If library_dir is specified, name must be specified too.
+          path (Path): repository's full filesystem path, which can be specified instead of name and library dir
+          urls (list): repository's urls
+          custom (dict): a dictionary containing other data
+          default_branch (str): repository's default branch ("main" if not defined)
+        """
+        if isinstance(library_dir, str):
+            library_dir = Path(library_dir)
+        if isinstance(path, str):
+            path = Path(path)
+
+        if (library_dir, name).count(None) == 1:
+            raise InvalidRepositoryError(
+                "Both library_dir and name need to be specified"
+            )
+
+        if name is not None:
+            self.name = self._validate_repo_name(name)
+            self.path = self._validate_repo_path(library_dir, name, path)
+            self.library_dir = library_dir.resolve()
+        elif path is None:
+            raise InvalidRepositoryError(
+                "Either speicfy library dir and name pair or path!"
+            )
+        else:
+            # maintain support for repositories whose names are not of significance
+            # in that case, only full path is specified (can happen if only using the GitRepository class)
+            # without the rest of the framework in some cotext
+            # name is still required for logging, so determine it based on the path
+            # if the path points to a direcotry directly inside the root direcotry
+            # set name to the name of that folder
+            # otherwise, use the same format that is expected when name is specified
+            if path.parent.parent != path.parent:
+                self.name = f"{path.parent.name}/{path.name}"
+                self.library_dir = path.parent.parent
             else:
-                repo_urls = [
-                    os.path.normpath(os.path.join(self.path, url))
-                    if not os.path.isabs(url)
-                    else url
-                    for url in repo_urls
-                ]
-        self.repo_urls = repo_urls
-        self.additional_info = additional_info
+                self.name = path.name
+                self.library_dir = path.parent
+            self.library_dir = self.library_dir.resolve()
+            self.path = self._validate_repo_path(path, None)
+
+        self.urls = self._validate_urls(urls)
+        self.default_branch = default_branch
+        self.custom = custom or {}
+
+    @classmethod
+    def from_json_dict(cls, json_data):
+        """Create a new instance based on data contained by the `json_data` dictionary,
+        which can be create by calling `to_json_dict`
+        """
+        return cls(**json_data)
+
+    def to_json_dict(self):
+        """Returns a dictionary mapping all attributes to their values"""
+        return {
+            "library_dir": str(self.library_dir),
+            "name": self.name,
+            "urls": self.urls,
+            "default_branch": self.default_branch,
+            "custom": self.custom,
+        }
 
     logging_functions = {
         logging.DEBUG: taf_logger.debug,
@@ -68,10 +111,6 @@ class GitRepository:
 
     log_template = "{}{}"
 
-    @property
-    def path(self):
-        return str(self._path)
-
     _remotes = None
 
     @property
@@ -83,7 +122,7 @@ class GitRepository:
     @property
     def is_git_repository_root(self):
         """Check if path is git repository."""
-        git_path = self._path / ".git"
+        git_path = self.path / ".git"
         return self.is_git_repository and (git_path.is_dir() or git_path.is_file())
 
     @property
@@ -110,9 +149,9 @@ class GitRepository:
 
     def _git(self, cmd, *args, **kwargs):
         """Call git commands in subprocess
-    e.g.:
-      self._git('checkout {}', branch_name)
-    """
+        e.g.:
+          self._git('checkout {}', branch_name)
+        """
         log_error = kwargs.pop("log_error", False)
         log_error_msg = kwargs.pop("log_error_msg", "")
         reraise_error = kwargs.pop("reraise_error", False)
@@ -125,12 +164,12 @@ class GitRepository:
         result = None
         if log_error or log_error_msg:
             try:
-                result = run(command)
+                result = run(command, **kwargs)
                 if log_success_msg:
                     self._log_debug(log_success_msg)
             except subprocess.CalledProcessError as e:
                 if error_if_not_exists and (
-                    not self._path.is_dir() or not self.is_git_repository
+                    not self.path.is_dir() or not self.is_git_repository
                 ):
                     log_error_msg = (
                         f"{self.path} does not exist or is not a git repository"
@@ -149,7 +188,7 @@ class GitRepository:
                     raise error
         else:
             try:
-                result = run(command)
+                result = run(command, **kwargs)
             except subprocess.CalledProcessError as e:
                 raise GitError(self, command=command, error=e)
             if log_success_msg:
@@ -180,7 +219,9 @@ class GitRepository:
         """
         if branch is None:
             branch = ""
-        commits = self._git("log {} --format=format:%H", branch, log_error=True).strip()
+        commits = self._git("log {} --format=format:%H", branch, log_error=True)
+        if commits is not None:
+            commits = commits.strip()
         if not commits:
             commits = []
         else:
@@ -199,9 +240,9 @@ class GitRepository:
             return self.all_commits_on_branch(branch=branch, reverse=reverse)
         if branch is None:
             branch = "HEAD"
-        commits = self._git(
-            "rev-list {}..{}", since_commit, branch, log_error=True
-        ).strip()
+        commits = self._git("rev-list {}..{}", since_commit, branch, log_error=True)
+        if commits is not None:
+            commits = commits.strip()
         if not commits:
             commits = []
         else:
@@ -214,7 +255,8 @@ class GitRepository:
         )
         return commits
 
-    def all_fetched_commits(self, branch="master"):
+    def all_fetched_commits(self, branch=None):
+        branch = branch or self.default_branch
         commits = self._git("rev-list ..origin/{}", branch).strip()
         if not commits:
             commits = []
@@ -256,7 +298,8 @@ class GitRepository:
         )
         if local_branches:
             local_branches = [
-                branch.replace("*", "").strip() for branch in local_branches
+                branch.replace("*", "").replace("+", "").strip()
+                for branch in local_branches
             ]
         remote_branches = self._git(
             f"branch -r --contains {commit}", log_error=True
@@ -298,11 +341,20 @@ class GitRepository:
                     log_error_msg=f"Repo {self.name}: could not list remote branches",
                     reraise_error=True,
                 )
-                # this command should return the branch's name if the remote branch
+                # this command should return the branch's name if the remote tracking branch
                 # exists
                 # it will also return some warnings if there are problems with some refs
                 if branch_name in remote_branch:
                     return True
+
+                # finally, check remote branch
+                if self.has_remote():
+                    return branch_name in self._git(
+                        f"ls-remote --heads origin {branch_name}",
+                        log_error_msg=f"Repo {self.name}: could check if the branch exists in the remote repository",
+                        reraise_error=True,
+                    )
+
         return False
 
     def branch_off_commit(self, branch_name, commit):
@@ -322,9 +374,9 @@ class GitRepository:
 
     def checkout_branch(self, branch_name, create=False, raise_anyway=False):
         """Check out the specified branch. If it does not exists and
-    the create parameter is set to True, create a new branch.
-    If the branch does not exist and create is set to False,
-    raise an exception."""
+        the create parameter is set to True, create a new branch.
+        If the branch does not exist and create is set to False,
+        raise an exception."""
         try:
             self._git(
                 "checkout {}",
@@ -336,7 +388,7 @@ class GitRepository:
                 raise (e)
 
             # skip worktree errors
-            if "is already checked out at" in e.error.output:
+            if "is already checked out at" in str(e):
                 return
 
             if create:
@@ -365,11 +417,10 @@ class GitRepository:
         self._git("clean -fd")
 
     def clone(self, no_checkout=False, bare=False, **kwargs):
-
         self._log_info("cloning repository")
         shutil.rmtree(self.path, True)
-        self._path.mkdir(exist_ok=True, parents=True)
-        if self.repo_urls is None:
+        self.path.mkdir(exist_ok=True, parents=True)
+        if self.urls is None:
             raise GitError(
                 repo=self, message="cannot clone repository. No urls were specified"
             )
@@ -388,7 +439,8 @@ class GitRepository:
         params = " ".join(params)
 
         cloned = False
-        for url in self.repo_urls:
+        for url in self.urls:
+            self._log_info(f"trying to clone from {url}")
             try:
                 self._git(
                     "clone {} . {}",
@@ -398,9 +450,10 @@ class GitRepository:
                     log_error_msg=f"cannot clone from url {url}",
                     reraise_error=True,
                 )
-            except GitError:
-                pass
+            except GitError as e:
+                self._log_info(f"could not clone from {url} due to {e}")
             else:
+                self._log_info(f"successfully cloned from {url}")
                 cloned = True
                 break
 
@@ -413,7 +466,7 @@ class GitRepository:
         Return old and new HEAD.
         """
         if branches is None:
-            branches = ["master"]
+            branches = [self.default_branch]
         self._log_debug(f"cloning or pulling branches {', '.join(branches)}")
 
         old_head = self.head_commit_sha()
@@ -465,9 +518,7 @@ class GitRepository:
         to check out previously checked out branch
         """
         if not self.branch_exists(branch_name, include_remotes=False):
-            current_branch = self.get_current_branch()
-            self.checkout_branch(branch_name)
-            self.checkout_branch(current_branch)
+            self._git(f"branch {branch_name} {self.remotes[0]}/{branch_name}")
 
     def checkout_commit(self, commit):
         self._git(
@@ -485,15 +536,24 @@ class GitRepository:
             self._git("diff --cached --exit-code --shortstat", reraise_error=True)
         except GitError:
             try:
-                run("git", "-C", self.path, "commit", "--quiet", "-m", message)
+                run("git", "-C", str(self.path), "commit", "--quiet", "-m", message)
             except subprocess.CalledProcessError as e:
                 raise GitError(
-                    repo=self, message=f"could not commit changes due to:\n{e.output}"
+                    repo=self, message=f"could not commit changes due to:\n{e}"
                 )
         return self._git("rev-parse HEAD")
 
     def commit_empty(self, message):
-        run("git", "-C", self.path, "commit", "--quiet", "--allow-empty", "-m", message)
+        run(
+            "git",
+            "-C",
+            str(self.path),
+            "commit",
+            "--quiet",
+            "--allow-empty",
+            "-m",
+            message,
+        )
         return self._git("rev-parse HEAD")
 
     def commits_on_branch_and_not_other(
@@ -502,7 +562,7 @@ class GitRepository:
         """
         Meant to find commits belonging to a branch which branches off of
         a commit from another branch. For example, to find only commits
-        on a speculative branch and not on the master branch.
+        on a speculative branch and not on the main branch.
         """
 
         self._log_debug(
@@ -518,11 +578,18 @@ class GitRepository:
         self._log_debug(f"found the following commits: {commits}")
         return commits
 
+    def commit_before_commit(self, commit):
+        return self._git(f"log {commit}^ -1 --pretty=%H", log_error=True).strip()
+
     def delete_local_branch(self, branch_name, force=False):
         """Deletes local branch."""
         flag = "-D" if force else "-d"
         self._git(f"branch {flag} {branch_name}", log_error=True)
 
+    def delete_remote_tracking_branch(self, remote_branch_name, force=False):
+        flag = "-D" if force else "-d"
+        self._git(f"branch {flag} -r {remote_branch_name}", log_error=True)
+
     def delete_remote_branch(self, branch_name, remote=None):
         if remote is None:
             remote = self.remotes[0]
@@ -536,13 +603,20 @@ class GitRepository:
         """Get commit sha of HEAD~{behind_head}"""
         return self._git("rev-parse HEAD~{}", behind_head)
 
-    def get_json(self, commit, path):
-        s = self.get_file(commit, path)
+    def get_json(self, commit, path, raw=False):
+        s = self.get_file(commit, path, raw=raw)
         return json.loads(s)
 
-    def get_file(self, commit, path):
+    def get_file(self, commit, path, raw=False):
         path = Path(path).as_posix()
-        return self._git("show {}:{}", commit, path)
+        return self._git("show {}:{}", commit, path, raw=raw)
+
+    def get_first_commit_on_branch(self, branch=None):
+        branch = branch or self.default_branch
+        first_commit = self._git(
+            f"rev-list --max-parents=0 {branch}", error_if_not_exists=False
+        )
+        return first_commit.strip() if first_commit else None
 
     def get_last_branch_by_committer_date(self):
         """Find the latest branch based on committer date. Should only be used for
@@ -582,23 +656,32 @@ class GitRepository:
                 branch = ""
             self._git("fetch {} {}", remote, branch, log_error=True)
 
+    def find_worktree_path_by_branch(self, branch_name):
+        """Returns path of the workree where the branch is checked out, or None if not checked out in any worktree"""
+        worktrees = self.list_worktrees()
+        for path, _, _branch_name in worktrees.values():
+            if _branch_name == branch_name:
+                return path
+        return None
+
     def get_current_branch(self):
         """Return current branch."""
         return self._git("rev-parse --abbrev-ref HEAD").strip()
 
-    def get_last_remote_commit(self, url, branch="master"):
+    def get_last_remote_commit(self, url, branch=None):
         """
-        Fet the last remote commit of the specified branch
+        Fetch the last remote commit of the specified branch
         """
-        if url is not None:
-            last_commit = self._git(
-                f"--no-pager ls-remote {url} {branch}", log_error=True
+        branch = branch or self.default_branch
+        if url is None:
+            raise FetchException(
+                "Could not fetch the last remote commit. URL not specified"
             )
-            if last_commit:
-                last_commit = last_commit.split("\t", 1)[0]
-                # in some cases (e.g. upstream is defined the result might contain a warning line)
-                return last_commit.split()[-1]
-        return None
+        last_commit = self._git(f"--no-pager ls-remote {url} {branch}", log_error=True)
+        if last_commit:
+            last_commit = last_commit.split("\t", 1)[0]
+            # in some cases (e.g. upstream is defined the result might contain a warning line)
+            return last_commit.split()[-1]
 
     def get_merge_base(self, branch1, branch2):
         """Finds the best common ancestor between two branches"""
@@ -617,12 +700,12 @@ class GitRepository:
             return None
 
     def init_repo(self, bare=False):
-        if self._path.is_dir():
-            self._path.mkdir(exist_ok=True, parents=True)
+        if self.path.is_dir():
+            self.path.mkdir(exist_ok=True, parents=True)
         flag = "--bare" if bare else ""
         self._git(f"init {flag}", error_if_not_exists=False)
-        if self.repo_urls is not None and len(self.repo_urls):
-            self._git("remote add origin {}", self.repo_urls[0])
+        if self.urls is not None and len(self.urls):
+            self._git("remote add origin {}", self.urls[0])
 
     def is_remote_branch(self, branch_name):
         for remote in self.remotes:
@@ -655,7 +738,8 @@ class GitRepository:
 
         return self._git("log {} {}", branch, " ".join(params)).split("\n")
 
-    def list_commit_shas(self, branch="master"):
+    def list_commit_shas(self, branch=None):
+        branch = branch or self.default_branch
         return self.list_commits(branch, format="format:%H")
 
     def list_n_commits(self, number=10, skip=None, branch=None):
@@ -695,6 +779,22 @@ class GitRepository:
             untracked_file for untracked_file in untracked_files if len(untracked_file)
         ]
 
+    def list_worktrees(self):
+        """
+        Returns a dictionary containing information about repository's worktrees:
+        {
+            "worktree1_path: (worktree1_path, worktree1_commit, worktree1_branch),
+            "worktree2_path: (worktree2_path, worktree2_commit, worktree2_branch),
+            ...
+        }
+        """
+        worktrees_list = self._git("worktree list")
+        worktrees = [w.split() for w in worktrees_list.splitlines() if w]
+        return {
+            Path(wt[0]): (Path(wt[0]), wt[1], wt[2].replace("[", "").replace("]", ""))
+            for wt in worktrees
+        }
+
     def merge_commit(self, commit):
         self._git("merge {}", commit, log_error=True)
 
@@ -746,26 +846,29 @@ class GitRepository:
         uncommitted_changes = self._git("status --porcelain")
         return bool(uncommitted_changes)
 
-    def synced_with_remote(self, branch="master", url=None):
+    def synced_with_remote(self, branch=None, url=None):
         """Checks if local branch is synced with its remote branch"""
         # check if the latest local commit matches
         # the latest remote commit on the specified branch
-        if url is None:
-            if self.repo_urls is not None and len(self.repo_urls):
-                url = self.repo_urls[0]
-            else:
-                url = self.get_remote_url()
+        urls = (
+            [url]
+            if url is not None
+            else self.urls
+            if self.urls
+            else [self.get_remote_url()]
+        )
+        for url in urls:
+            branch = branch or self.default_branch
+            tracking_branch = self.get_tracking_branch(branch, strip_remote=True)
+            if not tracking_branch:
+                return False
 
-        tracking_branch = self.get_tracking_branch(branch, strip_remote=True)
-        if not tracking_branch:
-            return False
-
-        try:
-            local_commit = self._git(f"rev-parse {branch}")
-        except GitError as e:
-            if "unknown revision or path not in the working tree" not in str(e):
-                raise e
-            local_commit = None
+            try:
+                local_commit = self._git(f"rev-parse {branch}")
+            except GitError as e:
+                if "unknown revision or path not in the working tree" not in str(e):
+                    raise e
+                local_commit = None
 
         remote_commit = self.get_last_remote_commit(url, tracking_branch)
 
@@ -774,8 +877,40 @@ class GitRepository:
     def top_commit_of_branch(self, branch):
         return self._git(f"rev-parse {branch}")
 
+    def _validate_repo_name(self, name):
+        """Ensure the repo name is not malicious"""
+        match = _repo_name_re.match(name)
+        if not match:
+            taf_logger.error(f"Repository name {name} is not valid")
+            raise InvalidRepositoryError(
+                "Repository name must be in format namespace/repository "
+                "and can only contain letters, numbers, underscores and "
+                f'dashes, but got "{name}"'
+            )
+        return name
+
+    def _validate_repo_path(self, library_dir, name, path=None):
+        """
+        validate repo path
+        (since this is coming from potentially untrusted data)
+        """
+        # using OS to avoid resolve adding drive letter name on Windows
+        repo_dir = os.path.join(str(library_dir), name or "")
+        repo_dir = os.path.normpath(repo_dir)
+        if not repo_dir.startswith(str(Path(library_dir))):
+            taf_logger.error(f"Repository path/name {library_dir}/{name} is not valid")
+            raise InvalidRepositoryError(
+                f"Repository path/name {library_dir}/{name} is not valid"
+            )
+        repo_dir = Path(repo_dir).resolve()
+        if path is not None and path != repo_dir:
+            raise InvalidRepositoryError(
+                "Both library dir and name pair and path specified and are not equal. Omit the path."
+            )
+        return repo_dir
+
     def _validate_url(self, url):
-        """ ensure valid URL """
+        """ensure valid URL"""
         for _url_re in [_http_fttp_url, _ssh_url]:
             match = _url_re.match(url)
             if match:
@@ -785,61 +920,21 @@ class GitRepository:
             f'Repository URL must be a valid URL, but got "{url}".'
         )
 
-
-class NamedGitRepository(GitRepository):
-    def __init__(
-        self,
-        root_dir,
-        repo_name,
-        repo_urls=None,
-        additional_info=None,
-        default_branch="master",
-        *args,
-        **kwargs,
-    ):
-        """
-    Args:
-      root_dir: the root directory
-      repo_name: repository's path relative to the root directory root_dir
-      repo_urls: repository's urls (optional)
-      additional_info: a dictionary containing other data (optional)
-      default_branch: repository's default branch
-    path is the absolute path to this repository. It is set by joining
-    root_dir and repo_name.
-    """
-        self.root_dir = root_dir
-        self.name = repo_name
-        path = self._get_repo_path(root_dir, repo_name)
-        super().__init__(
-            path,
-            repo_name=repo_name,
-            repo_urls=repo_urls,
-            additional_info=additional_info,
-            default_branch=default_branch,
-        )
-
-    def _get_repo_path(self, root_dir, repo_name):
-        """
-        get the path to a repo and ensure it is valid.
-        (since this is coming from potentially untrusted data)
-        """
-        self._validate_repo_name(repo_name)
-        repo_dir = str((Path(root_dir) / (repo_name or "")))
-        if not repo_dir.startswith(repo_dir):
-            self._log_error("repository name is not valid")
-            raise InvalidRepositoryError(f"Invalid repository name: {repo_name}")
-        return repo_dir
-
-    def _validate_repo_name(self, repo_name):
-        """ Ensure the repo name is not malicious """
-        match = _repo_name_re.match(repo_name)
-        if not match:
-            self._log_error("repository name is not valid")
-            raise InvalidRepositoryError(
-                "Repository name must be in format namespace/repository "
-                "and can only contain letters, numbers, underscores and "
-                f'dashes, but got "{repo_name}"'
-            )
+    def _validate_urls(self, urls):
+        if urls is not None:
+            if settings.update_from_filesystem is False:
+                for url in urls:
+                    self._validate_url(url)
+            else:
+                urls = [
+                    str((self.path / url).resolve())
+                    if (self.path / url).resolve().is_dir()
+                    else str((self.path).resolve())
+                    if not os.path.isabs(url)
+                    else url
+                    for url in urls
+                ]
+        return urls
 
 
 _repo_name_re = re.compile(r"^\w[\w_-]*/\w[\w_-]*$")
