@@ -34,6 +34,7 @@ from taf.exceptions import (
     TimestampMetadataUpdateError,
     YubikeyError,
     InvalidRepositoryError,
+    KeystoreError,
 )
 from taf.git import GitRepository
 from taf.utils import normalize_file_line_endings, on_rm_error
@@ -65,6 +66,15 @@ def is_delegated_role(role):
     return role not in ("root", "targets", "snapshot", "timestamp")
 
 
+def is_auth_repo(repo_path):
+    """Check if the given path contains a valid TUF repository"""
+    try:
+        Repository(repo_path)._repository
+        return True
+    except Exception:
+        return False
+
+
 def load_role_key(keystore, role, password=None, scheme=DEFAULT_RSA_SIGNATURE_SCHEME):
     """Loads the specified role's key from a keystore file.
     The keystore file can, but doesn't have to be password protected.
@@ -88,14 +98,17 @@ def load_role_key(keystore, role, password=None, scheme=DEFAULT_RSA_SIGNATURE_SC
         password = None
     key = role_keys_cache.get(role)
     if key is None:
-        if password is not None:
-            key = import_rsa_privatekey_from_file(
-                str(Path(keystore, role)), password, scheme=scheme
-            )
-        else:
-            key = import_rsa_privatekey_from_file(
-                str(Path(keystore, role)), scheme=scheme
-            )
+        try:
+            if password is not None:
+                key = import_rsa_privatekey_from_file(
+                    str(Path(keystore, role)), password, scheme=scheme
+                )
+            else:
+                key = import_rsa_privatekey_from_file(
+                    str(Path(keystore, role)), scheme=scheme
+                )
+        except FileNotFoundError:
+            raise KeystoreError(f"Cannot find {role} key in {keystore}")
     if not DISABLE_KEYS_CACHING:
         role_keys_cache[role] = key
     return key
@@ -158,22 +171,18 @@ def yubikey_signature_provider(name, key_id, key, data):  # pylint: disable=W061
 
 class Repository:
     def __init__(self, path, repo_name="default"):
-        self._path = Path(path)
+        self.path = Path(path)
         self.name = repo_name
 
     _framework_files = ["repositories.json", "test-auth-repo"]
 
-    @property
-    def path(self):
-        return str(self._path)
-
     @property
     def targets_path(self):
-        return self._path / TARGETS_DIRECTORY_NAME
+        return self.path / TARGETS_DIRECTORY_NAME
 
     @property
     def metadata_path(self):
-        return self._path / METADATA_DIRECTORY_NAME
+        return self.path / METADATA_DIRECTORY_NAME
 
     _tuf_repository = None
 
@@ -185,11 +194,11 @@ class Repository:
 
     @property
     def repo_id(self):
-        return GitRepository(self.path).initial_commit
+        return GitRepository(path=self.path).initial_commit
 
     @property
     def certs_dir(self):
-        certs_dir = self._path / "certs"
+        certs_dir = self.path / "certs"
         certs_dir.mkdir(parents=True, exist_ok=True)
         return str(certs_dir)
 
@@ -272,11 +281,21 @@ class Repository:
         """
         # before attempting to tuf repository, create empty targets directory if it does not exist
         # to avoid errors raised by tuf
+        targets_existed = True
         if not self.targets_path.is_dir():
+            targets_existed = False
             self.targets_path.mkdir(parents=True, exist_ok=True)
+        current_dir = self.metadata_path / "current"
+        previous_dir = self.metadata_path / "previous"
+        if current_dir.is_dir():
+            shutil.rmtree(current_dir)
+        if previous_dir.is_dir():
+            shutil.rmtree(previous_dir)
         try:
-            self._tuf_repository = load_repository(path, self.name)
+            self._tuf_repository = load_repository(str(path), self.name)
         except RepositoryError:
+            if not targets_existed:
+                self.targets_path.rmdir()
             raise InvalidRepositoryError(f"{self.name} is not a valid TUF repository!")
 
     def reload_tuf_repository(self):
