@@ -6,16 +6,23 @@ import logging
 # unusable.
 temporary_directory = None
 
+conf_directory_root = None
+
 update_from_filesystem = False
 
 validate_repo_name = True
 
+default_branch = "main"
+
 # Allows usage of specified commit as last validated commit
 # Useful when validating a local repository
 overwrite_last_validated_commit = False
 
 last_validated_commit = None
 
+# determines if script files will be loaded from disk
+development_mode = True
+
 # The 'log.py' module manages TUF's logging system.  Users have the option to
 # enable/disable logging to a file via 'ENABLE_FILE_LOGGING', or
 # tuf.log.enable_file_logging() and tuf.log.disable_file_logging().
