@@ -38,6 +38,15 @@ def attach_to_group(group):
         developer_tool.add_signing_key(path, role, pub_key_path, keystore,
                                        keys_description, scheme)
 
+    @metadata.command()
+    @click.argument("path")
+    @click.argument("keys-description")
+    @click.option("--keystore", default=None, help="Location of the keystore files")
+    @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme "
+                  "used for signing")
+    def add_roles(path, keystore, keys_description, scheme):
+        developer_tool.add_roles(path, keystore, keys_description, scheme)
+
     @metadata.command()
     @click.argument("path")
     @click.argument("role")
