@@ -1,7 +1,7 @@
 import click
 import taf.developer_tool as developer_tool
 from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
-from taf.updater.updater import update_repository, validate_repository
+from taf.updater.updater import update_repository, validate_repository, UpdateType
 
 
 def attach_to_group(group):
@@ -66,10 +66,10 @@ def attach_to_group(group):
 
     @repo.command()
     @click.argument("path")
-    @click.option("--root-dir", default=None, help="Directory where target repositories and, "
+    @click.option("--library-dir", default=None, help="Directory where target repositories and, "
                   "optionally, authentication repository are located. If omitted it is "
                   "calculated based on authentication repository's path. "
-                  "Authentication repo is presumed to be at root-dir/namespace/auth-repo-name")
+                  "Authentication repo is presumed to be at library-dir/namespace/auth-repo-name")
     @click.option("--namespace", default=None, help="Namespace of the target repositories. "
                   "If omitted, it will be assumed that namespace matches the name of the "
                   "directory which contains the authentication repository")
@@ -79,7 +79,7 @@ def attach_to_group(group):
     @click.option("--custom", default=None, help="A dictionary containing custom "
                   "targets info which will be added to repositories.json")
     @click.option("--use-mirrors", is_flag=True, help="Whether to generate mirrors.json or not")
-    def generate_repositories_json(path, root_dir, namespace, targets_rel_dir, custom, use_mirrors):
+    def generate_repositories_json(path, library_dir, namespace, targets_rel_dir, custom, use_mirrors):
         """
         Generate repositories.json. This file needs to be one of the authentication repository's
         target files or the updater won't be able to validate target repositories.
@@ -96,7 +96,7 @@ def attach_to_group(group):
         Authentication repository's namespace can, but does not have to be equal to the namespace of target,
         repositories. If the authentication repository's path is E:\\root\\namespace\\auth-repo, root
         directory will be determined as E:\\root. If this default value is not correct, it can be redefined
-        through the --root-dir option. If the --namespace option's value is not provided, it is assumed
+        through the --library-dir option. If the --namespace option's value is not provided, it is assumed
         that the namespace of target repositories is equal to the authentication repository's namespace,
         determined based on the repository's path. E.g. Namespace of E:\\root\\namespace2\\auth-repo
         is namespace2.
@@ -132,15 +132,15 @@ def attach_to_group(group):
         It is recommended that the content of the file is reviewed before doing so manually.
         """
         developer_tool.generate_repositories_json(
-            path, root_dir, namespace, targets_rel_dir, custom, use_mirrors
+            path, library_dir, namespace, targets_rel_dir, custom, use_mirrors
         )
 
     @repo.command()
     @click.argument("path")
-    @click.option("--root-dir", default=None, help="Directory where target repositories and, "
+    @click.option("--library-dir", default=None, help="Directory where target repositories and, "
                   "optionally, authentication repository are located. If omitted it is "
                   "calculated based on authentication repository's path. "
-                  "Authentication repo is presumed to be at root-dir/namespace/auth-repo-name")
+                  "Authentication repo is presumed to be at library-dir/namespace/auth-repo-name")
     @click.option("--namespace", default=None, help="Namespace of the target repositories. "
                   "If omitted, it will be assumed that namespace matches the name of the "
                   "directory which contains the authentication repository")
@@ -160,7 +160,7 @@ def attach_to_group(group):
     @click.option("--test", is_flag=True, default=False, help="Indicates if the created repository "
                   "is a test authentication repository")
     @click.option("--scheme", default=DEFAULT_RSA_SIGNATURE_SCHEME, help="A signature scheme used for signing.")
-    def initialize(path, root_dir, namespace, targets_rel_dir, custom, use_mirrors, add_branch, keystore,
+    def initialize(path, library_dir, namespace, targets_rel_dir, custom, use_mirrors, add_branch, keystore,
                    keys_description, commit, test, scheme):
         """
         \b
@@ -175,30 +175,39 @@ def attach_to_group(group):
         In order to have greater control over individual steps and be able to review files created
         in the initialization process, execute the mentioned commands separately.
         """
-        developer_tool.init_repo(path, root_dir, namespace, targets_rel_dir, custom, use_mirrors,
+        developer_tool.init_repo(path, library_dir, namespace, targets_rel_dir, custom, use_mirrors,
                                  add_branch, keystore, keys_description, commit, test, scheme)
 
     @repo.command()
     @click.argument("url")
     @click.argument("clients-auth-path")
-    @click.option("--clients-root-dir", default=None, help="Directory where target repositories and, "
+    @click.option("--clients-library-dir", default=None, help="Directory where target repositories and, "
                   "optionally, authentication repository are located. If omitted it is "
                   "calculated based on authentication repository's path. "
                   "Authentication repo is presumed to be at root-dir/namespace/auth-repo-name")
+    @click.option("--default-branch", default="main", help="Name of the default branch, like mian or master")
     @click.option("--from-fs", is_flag=True, default=False, help="Indicates if the we want to clone a "
                   "repository from the filesystem")
-    @click.option("--authenticate-test-repo", is_flag=True, help="Indicates that the authentication "
-                  "repository is a test repository")
-    def update(url, clients_auth_path, clients_root_dir, from_fs, authenticate_test_repo):
+    @click.option("--expected-repo-type", default="either", type=click.Choice(["test", "official", "either"]),
+                  help="Indicates expected authentication repository type - test or official. If type is set to either, "
+                  "the updater will not check the repository's type")
+    @click.option("--error-if-unauthenticated", is_flag=True, help="Raise an error if the repository allows "
+                  "unauthentiated commits and the updater detected authenticated commits newer than local "
+                  "head commit")
+    @click.option("--scripts-root-dir", default=None, help="Scripts root directory, which can be used to move scripts "
+                  "out of the authentication repository for testing purposes (avoid dirty index). Scripts will be expected "
+                  "to be located in scripts_root_dir/repo_name directory")
+    def update(url, clients_auth_path, clients_library_dir, default_branch, from_fs, expected_repo_type, error_if_unauthenticated,
+               scripts_root_dir):
         """
         Update and validate local authentication repository and target repositories. Remote
         authentication's repository url needs to be specified when calling this command. If the
         authentication repository and the target repositories are in the same root directory,
         locations of the target repositories are calculated based on the authentication repository's
         path. If that is not the case, it is necessary to redefine this default value using the
-        --clients-root-dir option. This means that if authentication repository's path is
+        --clients-library-dir option. This means that if authentication repository's path is
         E:\\root\\namespace\\auth-repo, it will be assumed that E:\\root is the root direcotry
-        if clients-root-dir is not specified.
+        if clients-library-dir is not specified.
         Names of target repositories (as defined in repositories.json) are appended to the root repository's
         path thus defining the location of each target repository. If names of target repositories
         are namespace/repo1, namespace/repo2 etc and the root directory is E:\\root, path of the target
@@ -209,21 +218,31 @@ def attach_to_group(group):
         the "test" target file), use --authenticate-test-repo flag. An error will be raised
         if this flag is omitted in the mentioned case. Do not use this flag when validating a non-test
         repository as that will also result in an error.
+
+        Some repositories can contain unauthenticated commits in-between two authenticated ones. The updater
+        will check existance and order of all authenticated commits and ignore unauthenticated. To raise
+        an error if there are new unauthenticated commits (newer than the last local commit),
+        error-if-unauthenticated option can be used.
+
+        Scripts root directory option can be used to move scripts out of the authentication repository for
+        testing purposes (avoid dirty index). Scripts will be expected  o be located in scripts_root_dir/repo_name directory
         """
-        update_repository(url, clients_auth_path, clients_root_dir, from_fs,
-                          authenticate_test_repo)
+        update_repository(url, clients_auth_path, clients_library_dir, default_branch, from_fs,
+                          UpdateType(expected_repo_type),
+                          error_if_unauthenticated=error_if_unauthenticated, scripts_root_dir=scripts_root_dir)
 
     @repo.command()
     @click.argument("clients-auth-path")
-    @click.option("--clients-root-dir", default=None, help="Directory where target repositories and, "
+    @click.option("--clients-library-dir", default=None, help="Directory where target repositories and, "
                   "optionally, authentication repository are located. If omitted it is "
                   "calculated based on authentication repository's path. "
-                  "Authentication repo is presumed to be at root-dir/namespace/auth-repo-name")
+                  "Authentication repo is presumed to be at library-dir/namespace/auth-repo-name")
+    @click.option("--default-branch", default="main", help="Name of the default branch, like mian or master")
     @click.option("--from-commit", default=None, help="First commit which should be validated.")
-    def validate(clients_auth_path, clients_root_dir, from_commit):
+    def validate(clients_auth_path, clients_library_dir, default_branch, from_commit):
         """
         Validates an authentication repository which is already on the file system
         and its target repositories (which are also expected to be on the file system).
         Does not clone repositories, fetch changes or merge commits.
         """
-        validate_repository(clients_auth_path, clients_root_dir, from_commit)
+        validate_repository(clients_auth_path, clients_library_dir, default_branch, from_commit)
